#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <signal.h>
#include <time.h>
#include <libsap.h>

#define SAPPLAY_VER "SAP player v0.2"

#define BUFFERSIZE 44100

static int soundoutputdes = -1;
static int currentSong = -1, songs = -1, defSong = -1;
static time_t starttime = 0;

static char *outputdevice = "/dev/dsp";
static char *filename = NULL;
static int silencetime = 3;

int keypressed()
{
	fd_set rfds;
	struct timeval tv;
	char c[32];
	int r;

	r = 0;
	while (1) {
		FD_ZERO(&rfds);
		FD_SET(0, &rfds);
		tv.tv_sec = 0;
		tv.tv_usec = 0;
		if (select(1, &rfds, NULL, NULL, &tv)) {
			read(1, c, 32);
			r = 1;
		}
		else return r;
	}
}

void opensoundoutput()
{
        int stereo, frequency, oss_format;
	
	stereo = 1; // channels - 1
	frequency = 44100;
	oss_format = AFMT_S16_LE;
	
	soundoutputdes = open(outputdevice, O_WRONLY);
	if (soundoutputdes<0) {
		printf("error opening output device %s\n", outputdevice);
		exit(1);
	}
	
        ioctl(soundoutputdes, SNDCTL_DSP_RESET, 0);
        ioctl(soundoutputdes, SNDCTL_DSP_SETFMT, &oss_format);
        ioctl(soundoutputdes, SNDCTL_DSP_SETFMT, &oss_format);
        ioctl(soundoutputdes, SNDCTL_DSP_STEREO, &stereo);
        ioctl(soundoutputdes, SNDCTL_DSP_SPEED, &frequency);
}

void closesoundoutput()
{
        ioctl(soundoutputdes, SNDCTL_DSP_RESET, 0);
	close(soundoutputdes);
}

void printprogressbar()
{
	char text[79];
	int textlen;

	static unsigned int i=0, j=0, k=0;
	static char textbuf[79];

	if (songs<2) strcpy(text, " --- playing");
	else sprintf(text, " --- playing song #%d - hit enter to play next song", currentSong);
	textlen = strlen(text);

	for (j=0, k=i; j<sizeof textbuf; j++, k=(k+1)%textlen) textbuf[j]=text[k];
	textbuf[sizeof textbuf - 1]='\0';
	
	printf("%c[%s]%c",13,textbuf,13);
	fflush(stdout);

	i=(i+1)%textlen;
}

void sigint_handler(int signum)
{
	time_t endtime = time(NULL);
	long runningtime = endtime - starttime;
	printf("%c[%d:%02d] decoding of \"%s\" finished.                                        \n",
	       13,(int)runningtime/60,(int)runningtime%60,filename);
	fflush(stdout);
	closesoundoutput();
	exit(1);
}

void options(int argc, char **argv)
{
	while (1) {
		int c;
		int option_index = 0;
		static struct option long_options[] =
 		{
			{"version", 0, 0, 'v'},
 			{"help",    0, 0, 'h'},
			{"song",    1, 0, 's'},
			{"device",  1, 0, 'd'},
			{"wait",    1, 0, 'w'},
			{0, 0, 0, 0}
		};

		c = getopt_long (argc, argv, "vhs:d:w:", long_options, &option_index);
		if (c == -1) break;

		switch (c) {
			case 'h':
				printf("usage: %s [-s songnum] [-d outdev] [-w seconds] [-h] [-v] file.sap\n", argv[0]);
				exit(1);
				
			case 'v':
				printf(SAPPLAY_VER " by Michal Kunikowski\nbased on " LIBSAP_INFO "\n");
				exit(0);

			case 's':
				currentSong = atoi(optarg);
				break;
			
			case 'd':
				outputdevice = optarg;
				break;

			case 'w':
				silencetime = atoi(optarg);
				break;
				
			default:
				exit(1);
			}
	}

	if (optind != argc - 1) {
		printf ("Try `%s -h' for more information\n", argv[0]);
		exit(1);
	}
	
	filename = argv[optind];
}


int main(int argc, char **argv)
{
	sapMUSICstrc *currentFile;
	short int playBuffer[BUFFERSIZE];
	int i;

	starttime = time(NULL);

	signal(SIGINT, &sigint_handler);
	
	options(argc, argv);
	
	printf(SAPPLAY_VER "\n\n");
	
	currentFile = sapLoadMusicFile( filename );
	if (currentFile==NULL) {
		printf("%s: error loading file %s\n", argv[0], filename);
		return 1;
	}

	songs = currentFile->numOfSongs;
	defSong = currentFile->defSong;

	if (currentSong < 0) currentSong = defSong;
	else  currentSong = (currentSong < songs-1) ? currentSong : songs-1;
	
	printf("FILE \"%s\"\n", filename);
	printf("%s", currentFile->commentBuffer);
	if (currentFile->numOfSongs >= 0) printf("SONGS %d\n", songs);
	if (currentFile->defSong >= 0) printf("DEFSONG %d\n", defSong);
	if (currentFile->isStereo) printf("STEREO\n");
	printf("\n");

	// if current song is not the default song
	if (currentSong!=currentFile->defSong) {
		sapRenderBuffer(playBuffer, 1); // bug workaroud
		sapPlaySong( currentSong );
	}

	if (currentSong < 0) currentSong = 0;

	opensoundoutput();

	while(1) {
		printprogressbar();
		sapRenderBuffer(playBuffer, BUFFERSIZE/2); // BUFFERSIZE/2 coz data stereo

		// silence detector - changes song if silence is longer than 'silencetime' seconds
		if (silencetime>0) {
			static long silencecounter = 0;
			int silence=1;
			for (i=0; i<BUFFERSIZE; i++) if ((playBuffer[i]&0xffff)!=SILENCE_WORD) silence=0;
			if (silence) silencecounter+=BUFFERSIZE; else silencecounter=0;
			if (silencecounter >= 44100*silencetime) {
				currentSong = (currentSong+1) % songs;
				sapPlaySong(currentSong);
			}
		}

		write(soundoutputdes, playBuffer, sizeof playBuffer);

		if (keypressed()) {
			printf("\033[A");
			currentSong = (currentSong+1) % songs;
			sapPlaySong(currentSong);
		}
	}
	
	return 0;
}
